# -*- coding: utf-8 -*-
def to_excel(df):
    lines = ["" for i in range(1000)]
    sep = "|"
    columns = df.columns
    for col in columns:
        i =0
        lines[i] += "%s%s%s"%(col, sep, sep)
        i+=1
        lines[i] += "nunique%s%s%s"%(sep, df[col].nunique(), sep)
        i+=1
        lines[i] += "n_na%s%s%s"%(sep, df[col].isnull().sum(), sep)
        if df[col].nunique() < 50:
            counts = df[col].value_counts(normalize=True)
            for k in counts.index:
                v = counts[k]
                v = "%.2f"%(v)
                v = v.replace(".", ",")
                i+=1
                lines[i] += "%s%s%s%s"%(k,sep, v,sep)
        while (i<len(lines)-1):
            i+=1
            lines[i]+="%s%s"%(sep, sep)
            
    return "\n".join(lines)

def analyse_df(df):
    import pandas as pd
    r = pd.DataFrame()
    for col in df.columns:
        try:
            r = r.append(analyse_column(df[col], col))
        except: pass

    return r
def analyse_column(X, nom_col="coucou"):
    """
    Calcul des statistiques descriptives sur une série (min, max, moyenne).

    Input:
        X : pandas series
            \-> supposed to be numeric
    Output: str

    """
    import pandas as pd
    try:
        df              = pd.DataFrame()
        df["nb_ligne"]  = pd.Series(X.shape[0])
        df["min"]       = X.min()
        df["max"]       = X.max()
        df["min_index"] = X.index.min()
        df["max_index"] = X.index.max()
        df["mean"]      = X.mean()
        df["median"]    = X.median()


        try     :df["var"]= X.var()
        except Exception as e  : print("pbm var : %s"%e)
        try     :df["ecart_type"]= X.std()
        except Exception as e  : print("pbm srtd : %s"%e)
        try     :df["EcAbs"]= EcAbs(X)
        except Exception as e  : print("pbm Ecabs : %s"%e)
        try     :df["Kolmogorov stat"], df["Kolmogorov pvalue"]= Kolmogorov(X)
        except Exception as e  : print("pbm Kolmogorov : %s"%e)
        try     : df["Indice_de_Gini"] = Indice_de_Gini(X)
        except Exception as e  : print("pbmm gini :  %s" %e)

    except Exception as e:
        print("analyse_column: pbm = {%s}%s" % (e, eol))
    df.index = [nom_col]

    return df
def calculate_unique_by_files(array_path: "tableau") -> "pd.df":
    """
    Pour chaque fichier passer en paramètre, calcul calculate_unique_null_not_null
    et renvoie une df avec toutes les infos.
    return : df
    """
    import pandas as pd
    result = False
    for path in array_path:
        try:
            df = calculate_unique_by_file(path)
            if isinstance(result, bool):
                result = df
            else:
                result = result.append(df)
        except Exception as e:
            print(path)
            print(e)
            df = False
            pass
    return result
def calculate_unique_by_file(csv_path: "str", **kwg) -> "pd.df":
    """
    Créer une df et Appelle calculate_unique_null_not_null dessus.
    return : df
    """
    import pandas as pd
    from jr_data_science.functions_pour_analyser_des_df import calculate_unique_null_not_null
    df = get_df(csv_path, **kwg)
    df = calculate_unique_null_not_null(df)
    df["file"] = csv_path
    df["column"] = df.index
    return df
def get_df(csv_path:"str", sep:"bool"=False, index_col:"bool"=False)->"df":
    """
    Renvoie une dataframe.
    """
    import pandas as pd
    from jr_data_science.functions_de_decouverte_de_fichiers import find_delimiter
    if not sep:
        sep = find_delimiter(csv_path)
    if not index_col:
        index_col = None
    df  = pd.DataFrame.from_csv(csv_path, index_col=index_col, sep=sep)
    return df
def tendances_centrales(Y, eol="\n"):
    """
    Retourne une chaine de caractère décrivant la tendance centrale d'une série

    Input : pandas series

    Ouput : string
    """
    moyenne, mediane = Y.mean(),  Y.median()

    try:
        mode =  Y.mode()
    except:
        mode = False

    return moyenne, mediane , mode
def dispersion(X, eol="\n"):
    """
    Retourne une chaine de caractère décrivant la dispersion d'une série

    Input : pandas series

    Ouput : string

    """

    to_print  = "Dispersion%s"%eol
    to_print  += "Variance : %.2f%s"%(X.var(), eol)
    to_print  += "Ecart type : %.2f%s"%(X.std(), eol)
    to_print  += "Etendue : [%.2f, %.2f]%s"%(X.min(), X.max(), eol)
    to_print  += "Ecart absolu moyen : %s %s"%(EcAbs(X), eol)
    return to_print
def Indice_de_Gini(Y):
    """
    Retourne une chaine de caractère décrivant l'indice de gini' d'une série (concentration)

    Input : pandas series

    Ouput : string

    """

    return gini(Y)
def gini(list_of_values):
    """
    Calcul l'indice de gini d'une série

    Input : pandas series

    Ouput : float

    """
    sorted_list = sorted(list_of_values)
    height, area = 0, 0
    for value in sorted_list:
        height += value
        area += height - value / 2.
        fair_area = height * len(list_of_values) / 2

    return (fair_area - area) / float(fair_area)
def Kolmogorov(X):
    """
    Calcul le test de Kolmogorov

    Input : pandas series

    Ouput : string

    """
    from scipy import stats
    r = stats.kstest(X, 'norm')
    print (r)
    return r
def EcAbs(L):
    """
    Calcul de l'écart absolu moyen
    L : np.ndarray
    Ouput : string
    """
    import numpy as np
    if isinstance(L, np.ndarray) :
        L[np.isnan(L)] = L[~np.isnan(L)].mean()
    m = L.sum() / len(L)
    M = [abs(x - m) for x in L]
    EcAbs_ = sum(M, 0.0) / len(M)

    return EcAbs_
def smooth(x,window_len=125,window='hamming'):
    """
    Smooth a serie thanks to hamming method.
    """
    if x.ndim != 1:
            raise Exception(ValueError, "smooth only accepts 1 dimension arrays.")
    if x.size < window_len:
            raise Exception(ValueError, "Input vector needs to be bigger than window size.")
    if window_len<3:
            return x
    if not window in ['flat', 'hanning', 'hamming', 'bartlett', 'blackman']:
            raise Exception(ValueError, "Window is on of 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'")
    s=np.r_[2*x[0]-x[window_len-1::-1],x,2*x[-1]-x[-1:-window_len:-1]]
    if window == 'flat': #moving average
            w=np.ones(window_len,'d')
    else:
            w=eval('np.'+window+'(window_len)')
    y=np.convolve(w/w.sum(),s,mode='same')
    return y[window_len:-window_len+1]
def MSE(target, predictions):
    """
    Test predictor quality with Mean Square Error
    input :
        target : array of real values
        prediction : array of predicted values
    """
    squared_deviation = np.power(target-predictions, 2)
    return np.mean(squared_deviation)
def MAD(target, predictions):
    """
    Function to test predictor quality : Return Mean Average Deviation
    input :
        target : array of real values
        prediction : array of predicted values
    """
    absolute_deviation = np.abs(target-predictions)
    return np.mean(absolute_deviation)
def proba(ar):
    """
    Calculate the probability of each element in an array.
    Param :
        ar : array of value
    Return :
        dictionnary : [value : proba]

    """
    values = set(ar)
    nb = len(ar)
    print(nb)
    proba = [ar.count(v)/nb for v in values]
    return dict(list(zip(values, proba)))
def RunLinearModel(data, cible, temps ):
    import statsmodels.formula.api as smf
    import numpy as np
    import pandas
    print(cible in data.columns)
    print(temps in data.columns)
    requete = "%s ~ %s"%(cible, temps)
    model = smf.ols(requete, data=data)
    results = model.fit()
    return model, results
def get_needed_sample_size(original_pop, accepted_error=0.05):
    """
    Calculate Bienayme tchebiechef minimum sample size to get a defined acceptable error rate out of an original population size.
    inputs :
        original_pop : size of the pop
        accepted_error : max error we want to have
    """
    a = 1.96**2
    b = accepted_error**2
    c = a * original_pop
    e = original_pop-1
    d = a+b*e
    return  c/d
def get_confidence(original_pop, sample_size):
    """
    According to a reference population size and a test sample size, return the confidence we can have on the result.

    inputs :
        original_pop : size of the pop

    """

    from math import sqrt
    x = 1.96**2
    a = x * original_pop
    b = (original_pop-1) * sample_size * x
    c = float(a)/b
    d = sqrt(c)
    return  float(d)
def min_max(time_serie):
    """
    return a tuple with the min and max of time serie index
    """
    return (time_serie.index.min(), time_serie.index.max())
def reject_outliers(data, m=2.):
    import numpy as np
    d = np.abs(data - np.median(data))
    mdev = np.median(d)
    s = d / mdev if mdev else 0.
    return data[s < m]
def get_outliers(data, m=2.):
    import numpy as np
    d = np.abs(data - np.median(data))
    mdev = np.median(d)
    s = d / mdev if mdev else 0.
    return data[s > m]
def frange(start, stop, step):
    """
    Create an array containing numbers from [start] to [stop] with a step of [step]

    Parameters:
        start: int
        stop : int
        step : int

    Return:
        r : array of int
    """
    r = [start]
    i = start
    while i < stop:
        i += step
        r.extend([i])
    return r



def get_n_indices_set(df, n,  classe, colonne = -1):
    """
        Create an array of len(n) dataset containing lines number (ie defining datasets)
        
        df      : serie
        n       : number of wanted datasets
        classe  : value wanted in the serie
        colonne : useless ?
        
        Return a shuffled array of indices.
    """
    import random
    import numpy as np
    try:
        datasets = []
        for i in range(n):
            indices_classe = df[df==classe ].index 
            indices_classe = indices_classe.values 
            
            not_in_classe  = df[df!=classe ].index 
            not_in_classe  = not_in_classe.values  
            
            random.shuffle(indices_classe)
            random.shuffle(not_in_classe )                          
            
            smallest_n     = min(len(indices_classe), len(not_in_classe))
            indices        = np.concatenate([indices_classe[:smallest_n], not_in_classe[:smallest_n]])
            
            if smallest_n > 5:
                datasets.append(indices)
    except Exception as e :
        print("pbm in get_n_indices_set : %s"%e)
    return datasets
"""
    SVM grid search
"""
def svm_grid_search(df, Y):
    import sys
    from sklearn                   import grid_search
    from sklearn.svm               import SVC
    import pickle
    parameters = {  "C"              : [1.0, 0.5, 0.1, 2]                , 
                    "kernel"         : ['rbf', 'poly', 'sigmoid']        ,
                    "degree"         : [2, 3, 4,5,6]                     , 
                    "gamma"          : [0.01, 0.1, 0.001, 0.5]           , 
                    "coef0"          : [0.0]                             , 
                    "shrinking"      : [True]                            , 
                    "probability"    : [True]                            , 
                    "tol"            : [0.001]                           , 
                    "cache_size"     : [10]                              , 
                    "class_weight"   : [None]                            , 
                    "verbose"        : [False]                           , 
                    "max_iter"       : [-1]                              , 
                    "random_state"   : [None]                            }
    
    print("-"*30, "\n svm_grid_search"); sys.stdout.flush()
    
    """
        Search Grid
    """
    grid    = grid_search.GridSearchCV(SVC(), parameters, verbose=0, scoring='f1_weighted')
    try :grid.fit(df, Y)
    except Exception as e : print("[svm_grid_search - 2] : %s"%e) ; print("x_train, y_train : ", x_train, y_train)

    print(" best_score_ = %.2f with =  %s "%(grid.best_score_, grid.best_estimator_ )) ; sys.stdout.flush()    
    return grid.best_estimator_
"""
    KNN grid search
"""
def knn_grid_search(df, Y):
    import sys
    from sklearn                   import grid_search
    from sklearn.neighbors         import KNeighborsClassifier
    
    import pickle

    parameters = {  "n_neighbors"    : [2, 3, 5, 8 , 10]                         , 
                    "weights"         : ['uniform', 'distance']                  ,
                    "algorithm"      : ['auto', 'ball_tree', 'kd_tree', 'brute'] , 
                    "leaf_size"      : [1,3 ]                                    }
                    
    grid_type = "knn_grid_search"
    print("-"*30, "\n %s"%grid_type); sys.stdout.flush()
    """
        Search Grid
    """
    grid    = grid_search.GridSearchCV(KNeighborsClassifier(), parameters, verbose=0, scoring='f1_weighted')
    try :grid.fit(df, Y)
    except Exception as e : print("[%s - 2] : %s"%(grid_type,e)) ; print("x_train, y_train : ", x_train, y_train)

    print(" best_score_ = %.2f with =  %s "%(grid.best_score_, grid.best_estimator_ )) ; sys.stdout.flush()    
    return grid.best_estimator_
"""
    Random Forest
"""
def RandomForest_grid_search(df, Y):
    import sys, pickle
    from sklearn.ensemble          import RandomForestClassifier
    from sklearn                   import grid_search
    from sklearn.metrics           import make_scorer
    from sklearn.preprocessing     import label_binarize
    from sklearn.multiclass        import OneVsRestClassifier
    #grid search :
    parameters_nas1 = {  "n_estimators"    : [ 1000]                 , 
                    "min_samples_leaf"     : [3]                     ,
                    "criterion"            : ["gini", "entropy"]     ,
                    "max_features"         : ["auto"]                ,
                    "max_depth"            : [None]                  , 
                    "n_jobs"               : [ -1]                   ,  
                    "class_weight"         : ["auto"]                ,
                    "warm_start"           : [False]                 }
    
    parameters = {  "n_estimators"    : [ 500, 1000,2000, ]          , 
                    "min_samples_leaf"     : [3,1, 7 ]               ,
                    "criterion"            : ["gini", "entropy"]     ,
                    "max_features"         : ["auto"]                ,
                    "max_depth"            : [None]                  , 
                    "n_jobs"               : [ -1]                   }#,  
                    #"class_weight"         : ["auto"]          }
                    #"warm_start"           : [False]           }
    grid_type = "RandomForestClassifier"
    
    print("-"*30, "\n %s"%grid_type); sys.stdout.flush()
    """
        Search 
    """
    #model_to_set = OneVsRestClassifier(RandomForestClassifier())
    #grid         = grid_search.GridSearchCV(model_to_set, parameters, verbose=1, scoring='r2')

    grid         = grid_search.GridSearchCV(RandomForestClassifier(), parameters, verbose=1, scoring='f1')
    
    try :grid.fit(df, Y)
    except Exception as e : print("[%s - 2] : %s"%(grid_type,e)) 
    
    print(" best_score_ %s with =  %s "%( grid.best_score_,  grid.best_estimator_ )) ; sys.stdout.flush()    
    return grid.best_estimator_
"""
    Affinity Propagation
"""
def Affinityscorer(estimator, X, y):
    return float(sum(estimator.predict(X)==y))/len(y)
def AffinityPropagation_grid_search(df, Y):
    import sys, pickle
    from sklearn.cross_validation   import cross_val_score, train_test_split
    from sklearn                    import grid_search
    from sklearn.cluster            import AffinityPropagation
    from sklearn.metrics            import make_scorer
    
    parameters = {  "damping"    : [0.5,0.7,0.8,0.9]                       , 
                    "convergence_iter"         : [5,10,15]  ,
                    "max_iter"      : [10,20,50,100,150,200]    }
                    
    grid_type = "AffinityPropagation"
    print("-"*30, "\n %s"%grid_type); sys.stdout.flush()

    """
        Search 
    """
    scorer  = make_scorer(Affinityscorer)
    grid    = grid_search.GridSearchCV(AffinityPropagation(), parameters, verbose=0, scoring=scorer)

    
    try :
        grid.fit(df, Y)
        print("fitted"); sys.stdout.flush()
    except Exception as e : print("[%s - 2] : %s"%(grid_type,e)) 
    
        
    print(" best_score_ %s with =  %s "%( grid.best_score_,  grid.best_estimator_ )) ; sys.stdout.flush()    
    return grid.best_estimator_
def draw_confusion_matrix(y_test, predictions, title):
    cm    = confusion_matrix(y_test, predictions)
    accur = accuracy_score  (y_test, predictions)
    
    plt.matshow  (cm)
    plt.ylabel   ('True Label')
    plt.xlabel   ('Predicted Label')
    plt.suptitle ('Confusion matrix (Accuracy of %.2f) for [%s]'%(accur,title))
    plt.colorbar ()
    plt.show     ()
def draw_roc_curve(fitted_c, x_test, y_test, title):
    from sklearn.metrics import roc_curve, auc
    c                                        = fitted_c
    probas                                   = c.predict_proba(x_test)
    false_positive_rate, recall_, thresholds = roc_curve(y_test, probas[:,1])
    roc_auc                                  = auc(false_positive_rate, recall_)
    
    plt.title  ('ROC %.2f %s'%(roc_auc, title))
    plt.legend (loc="lower right")
    plt.plot   ([0,1],[0,1], "r--")
    plt.plot   (false_positive_rate, recall_, 'b', label='AUC = %.2f'%roc_auc)
    plt.xlim   ([0.0,1.1])
    plt.ylim   ([0.0,1.1])
    plt.ylabel ('Recall')
    plt.xlabel ('Fall-out')
    plt.show()
def multi_class_roc(c, X, y, classes, title):
    # imports 
    import numpy                  as     np
    import matplotlib.pyplot      as     plt
    
    from sklearn                  import svm, datasets
    from sklearn.metrics          import roc_curve, auc
    from sklearn.cross_validation import train_test_split
    from sklearn.preprocessing    import label_binarize
    from sklearn.multiclass       import OneVsRestClassifier
    # Binarize the output
    y         = label_binarize(y, classes=classes)
    n_classes = y.shape[1]
    X         = np.c_[X]

    # shuffle and split training and test sets
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=.2,random_state=0)

    # Learn to predict each class against the other
    classifier = OneVsRestClassifier(c)
    y_score    = classifier.fit(X_train, y_train).predict(X_test)

    # Compute ROC curve and ROC area for each class
    fpr = dict()
    tpr = dict()
    roc_auc = dict()
    for i in range(n_classes):
        fpr[i], tpr[i], _ = roc_curve(y_test[:, i], y_score[:, i])
        roc_auc[i] = auc(fpr[i], tpr[i])

    # Compute micro-average ROC curve and ROC area
    fpr["micro"], tpr["micro"], _ = roc_curve(y_test.ravel(), y_score.ravel())
    roc_auc["micro"]              = auc(fpr["micro"], tpr["micro"])

    # Plot ROC curve
    plt.figure()
    plt.plot(fpr["micro"], 
             tpr["micro"],
             label='micro-average ROC curve (area = {0:0.2f})'
                   ''.format(roc_auc["micro"]))
    for i in range(n_classes):
        plt.plot(fpr[i], tpr[i], label='ROC curve of class {0} (area = {1:0.2f})'
                                       ''.format(i, roc_auc[i]))

    plt.plot   ( [ 0   , 1   ], [0, 1], 'k--')
    plt.xlim   ( [ 0.0 , 1.0 ] )
    plt.ylim   ( [ 0.0 , 1.05] )
    plt.xlabel ( 'False Positive Rate')
    plt.ylabel ( 'True Positive Rate')
    plt.title  ( title)
    plt.legend ( loc="lower right")
    plt.show   ( )
def binary_multi_roc_curve(c, X, y, t ):
    import numpy                    as      np
    from   scipy                    import interp
    import matplotlib.pyplot        as     plt
    from   sklearn                  import svm, datasets
    from   sklearn.metrics          import roc_curve, auc
    from   sklearn.cross_validation import StratifiedKFold
    
    
    X.index = list(range(0, X.shape[0]))
    y.index = list(range(0, y.shape[0]))
    n_samples, n_features = X.shape
    X          = np.c_[X]
    cv         = StratifiedKFold(y, n_folds=6)
    classifier = c
    mean_tpr   = 0.0
    mean_fpr   = np.linspace(0, 1, 100)
    all_tpr    = []
    
    plt.figure(figsize=(10,5))
    plt.xlim([-0.05, 1.05])                       ; plt.xlabel('False Positive Rate')
    plt.ylim([-0.05, 1.05])                       ; plt.ylabel('True Positive Rate')
    plt.title('binary_multi_roc_curve for predicting [%s]'%t) ; plt.legend(loc="best")
    plt.plot([0, 1], [0, 1], '--', color=(0.6, 0.6, 0.6), label='Luck')
    
    for i, (train, test) in enumerate(cv):
        #cross prediction
        probas_              = classifier.fit(X[train], y[train]).predict_proba(X[test])
        fpr, tpr, thresholds = roc_curve(y[test], probas_[:, 1])
        #means
        mean_tpr            += interp(mean_fpr, fpr, tpr)
        mean_tpr[0]          = 0.0
        #ROC 
        roc_auc              = auc(fpr, tpr)
        #plot ROC
        plt.plot(fpr, tpr, lw=1, label='ROC fold %d (area = %0.2f)' % (i, roc_auc))
    
    #mean accuracy
    mean_tpr /= len(cv); mean_tpr[-1] = 1.0; mean_auc = auc(mean_fpr, mean_tpr)
    plt.plot(mean_fpr, mean_tpr, 'r--',label='Mean ROC (area = %0.2f)' % mean_auc, lw=2)
    plt.show()
    
        